import numpy as np
import pandas as pd


def corr_without_symmetry(df: pd.DataFrame) -> pd.DataFrame:
    corr = df.corr()
    return corr.mask(np.tril(np.ones(corr.shape)).astype(np.bool_))


def unroll(df: pd.DataFrame, datetime: pd.DatetimeIndex) -> pd.Series:
    df = df.melt(ignore_index=False).dropna()

    df.index = df.index + "_" + df.variable
    df = df.drop(columns=["variable"])
    df = df.rename(columns={"value": datetime})
    return df.squeeze()



from typing import TYPE_CHECKING, Dict, List, Optional

from krisi.evaluate.type import ScoreCardMetadata
from krisi.report.type import InteractiveFigure, PlotlyInput
from krisi.utils.environment import is_notebook
from krisi.utils.iterable_helpers import del_dict_keys, flatten

if TYPE_CHECKING:
    from dash import dcc, html


external_script = ["https://tailwindcss.com/", {"src": "https://cdn.tailwindcss.com"}]


def block(
    graph: "dcc.Graph",
    title: Optional["html.P"] = None,
    controllers: Optional["html.Div"] = None,
) -> "html.Div":
    return html.Div(
        children=[graph, title, controllers],
        className="flex flex-row flex-wrap",
    )


def figure_with_controller(figure: InteractiveFigure):
    if len(figure.inputs) > 0 or len(figure.global_input_ids) > 0:
        return block(
            graph=dcc.Graph(
                id=figure.id,
                figure=figure.get_figure(
                    width=figure.plot_args["width"] - 216.0,
                    **del_dict_keys(figure.plot_args, "width"),
                ),
                className="h-full flex align-center",
            ),
            title=None,
            controllers=html.Div(
                className="h-full flex align-center",
                children=[
                    input_.type(
                        className="h-full min-w-[150px] flex justify-center align-center",
                        id=input_.id,
                        options=input_.options,
                        value=input_.default_value,
                        clearable=False,
                    )
                    for input_ in figure.inputs
                ],
            ),
        )

    else:
        return dcc.Graph(
            className="h-full flex align-center",
            id=figure.id,
            figure=figure.get_figure(
                width=figure.plot_args["width"] - 216.0,
                **del_dict_keys(figure.plot_args, "width"),
            ),
            style={"display": "inline-block"},
        )


def category_block(category: str, figures: List[InteractiveFigure]) -> "html.Div":
    return html.Div(
        className="flex flex-col shadow-lg mb-4",
        children=[
            html.Div(
                className="flex items-center w-full h-12 bg-yellow-400 ",
                children=[
                    html.H2(
                        category,
                        className="flex m-0 items-center text-lg font-normal text-white leading-normal mt-0 mb-2 pl-6",
                    )
                ],
            ),
            html.Div(
                className="flex flex-wrap flex-row  p-6",
                children=[
                    *[figure_with_controller(figure) for figure in figures],
                ],
            ),
        ],
    )


def global_input_controller_block(input_):
    return input_.type(
        className="h-full min-w-[150px] flex justify-center align-center",
        id=input_.id,
        options=input_.options,
        value=input_.default_value,
        clearable=False,
    )


def name_description_block(name: str, description: str) -> "html.Div":
    return html.Div(
        className="flex flex-col h-full w-full p-2",
        children=[
            html.H3(name, className="flex-auto text-xl font-semibold text-slate-600"),
            html.P(description),
        ],
    )


def create_description_component(
    scorecard_metadata: Optional[ScoreCardMetadata],
) -> Optional["html.Div"]:
    if scorecard_metadata is not None:
        return html.Div(
            className="flex flex-row w-full h-20 mb-12",
            children=[
                name_description_block(
                    scorecard_metadata.project_name,
                    scorecard_metadata.project_description,
                ),
                name_description_block(
                    scorecard_metadata.model_name, scorecard_metadata.model_description
                ),
                name_description_block(
                    scorecard_metadata.dataset_name,
                    scorecard_metadata.dataset_description,
                ),
            ],
        )
    else:
        return None


def run_app(
    components: Dict[str, List[InteractiveFigure]],
    global_controllers: List[PlotlyInput],
    title: str = "",
    description: str = "",
    scorecard_metadata: Optional[ScoreCardMetadata] = None,
) -> None:
    import sys

    from dash import Dash, Input, Output, dcc, html
    from jupyter_dash import JupyterDash

    isnotebook = is_notebook()

    sys.modules[__name__].html = html
    sys.modules[__name__].dcc = dcc

    if isnotebook:
        app = JupyterDash(__name__, external_scripts=external_script)
    else:
        app = Dash(__name__, external_scripts=external_script)
    app.scripts.config.serve_locally = True

    app.layout = html.Div(
        className="p-24 w-full h-full flex flex-col",
        children=[
            html.H1(
                title,
                className="flex-auto text-3xl font-semibold text-yellow-400 p-2",
            ),
            create_description_component(scorecard_metadata),
            html.Div(
                children=[
                    global_input_controller_block(input_)
                    for input_ in global_controllers
                ]
            ),
            html.Div(
                className="flex flex-wrap flex-col",
                children=[
                    *[
                        category_block(category, figures)
                        for category, figures in components.items()
                        if len(figures) > 0
                    ],
                ],
            ),
        ],
    )

    for component in flatten(list(components.values())):
        if len(component.inputs) > 0 or len(component.global_input_ids) > 0:
            app.callback(
                Output(component.id, "figure"),
                [Input(input_.id, input_.value_name) for input_ in component.inputs]
                + [Input(input_id, "value") for input_id in component.global_input_ids],
            )(component.get_figure)

    if isnotebook:
        app.run_server(mode="inline", debug=False, threaded=True)
    else:
        app.run(debug=False, threaded=True)



import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    brier_score_loss,
    f1_score,
    matthews_corrcoef,
    precision_score,
    recall_score,
)

from krisi.evaluate.library.diagrams import (
    callibration_plot,
    display_single_value,
    display_time_series,
)
from krisi.evaluate.library.metric_wrappers import brier_multi
from krisi.evaluate.metric import Metric
from krisi.evaluate.type import MetricCategories

accuracy = Metric[float](
    name="Accuracy",
    key="accuracy",
    category=MetricCategories.class_err,
    info="In multilabel classification, this function computes subset accuracy: the set of labels predicted for a sample must exactly match the corresponding set of labels in y_true. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html",
    func=accuracy_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
)
""" ~ """

recall = Metric[float](
    name="Recall",
    key="recall",
    category=MetricCategories.class_err,
    info="The recall is the ratio tp / (tp + fn) where tp is the number of true positives and fn the number of false negatives. The recall is intuitively the ability of the classifier to find all the positive samples.\nThe best value is 1 and the worst value is 0. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
    parameters={"average": "binary"},
    func=recall_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
)
"""~"""
precision = Metric[float](
    name="Precision",
    key="precision",
    category=MetricCategories.class_err,
    info="The precision is the ratio tp / (tp + fp) where tp is the number of true positives and fp the number of false positives. The precision is intuitively the ability of the classifier not to label as positive a sample that is negative.\nThe best value is 1 and the worst value is 0. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html",
    parameters={"average": "binary"},
    func=precision_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
)
"""~"""
f_one_score = Metric[float](
    name="F1 Score",
    key="f_one_score",
    category=MetricCategories.class_err,
    info="The F1 score can be interpreted as a harmonic mean of the precision and recall, where an F1 score reaches its best value at 1 and worst score at 0. The relative contribution of precision and recall to the F1 score are equal. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html",
    parameters={"average": "macro"},
    func=f1_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
    supports_multiclass=True,
)
"""~"""
matthew_corr = Metric[float](
    name="Matthew Correlation Coefficient",
    key="matthew_corr",
    category=MetricCategories.class_err,
    info="The Matthews correlation coefficient is used in machine learning as a measure of the quality of binary and multiclass classifications. It takes into account true and false positives and negatives and is generally regarded as a balanced measure which can be used even if the classes are of very different sizes. The MCC is in essence a correlation coefficient value between -1 and +1. A coefficient of +1 represents a perfect prediction, 0 an average random prediction and -1 an inverse prediction. The statistic is also known as the phi coefficient.",
    func=matthews_corrcoef,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
)
"""~"""
brier_score = Metric[float](
    name="Brier Score",
    key="brier_score",
    category=MetricCategories.class_err,
    info="The smaller the Brier score loss, the better, hence the naming with “loss”. The Brier score measures the mean squared difference between the predicted probability and the actual outcome. The Brier score always takes on a value between zero and one, since this is the largest possible difference between a predicted probability (which must be between zero and one) and the actual outcome (which can take on values of only 0 and 1). It can be decomposed as the sum of refinement loss and calibration loss.",
    func=lambda y, pred, prob, **kwargs: brier_score_loss(
        y_true=y, y_prob=prob, **kwargs
    ),
    parameters=dict(pos_label=1),
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
)
"""~"""
calibration = Metric[float](
    name="Calibration Plot",
    key="calibration",
    category=MetricCategories.class_err,
    info="Used to plot the calibration of a model with it probabilities.",
    func=lambda y, pred, prob: pd.concat(
        [y, prob.iloc[:, 0].rename("probs")], axis="columns"
    ),
    plot_funcs=[(callibration_plot, dict(width=1500.0, bin_size=0.1))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
    supports_multiclass=True,
)
"""~"""

brier_score_multi = Metric[float](
    name="Brier Score Multilabel",
    key="brier_score_multi",
    category=MetricCategories.class_err,
    info="Multilabel calculation of the Brier score loss. The smaller the Brier score loss, the better, hence the naming with “loss”. The Brier score measures the mean squared difference between the predicted probability and the actual outcome. The Brier score always takes on a value between zero and one, since this is the largest possible difference between a predicted probability (which must be between zero and one) and the actual outcome (which can take on values of only 0 and 1). It can be decomposed as the sum of refinement loss and calibration loss.",
    func=lambda y, pred, prob, **kwargs: brier_multi(y, prob, **kwargs),
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_func_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
    supports_multiclass=True,
)
"""~"""
all_classification_metrics = [
    accuracy,
    recall,
    precision,
    f_one_score,
    matthew_corr,
    brier_score,
    brier_score_multi,
    calibration,
]
"""~"""
minimal_classification_metrics = [accuracy, f_one_score]
"""~"""
multilabel_classification = [
    metric for metric in all_classification_metrics if metric.supports_multiclass
]
"""~"""



import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Callable, List, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

MetricResult = TypeVar(
    "MetricResult", bound=Union[float, int, str, List, Tuple, pd.DataFrame, pd.Series]
)


Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
Probabilities = Union[np.ndarray, pd.Series, pd.DataFrame, List[Union[int, float]]]
PredictionsDS = pd.Series
ProbabilitiesDS = pd.DataFrame
TargetsDS = pd.Series
MetricFunctionProb = Callable[[PredictionsDS, TargetsDS, ProbabilitiesDS], MetricResult]
MetricFunctionNoProb = Callable[
    [PredictionsDS, TargetsDS, ProbabilitiesDS], MetricResult
]
MetricFunction = Union[MetricFunctionNoProb, MetricFunctionProb]


class MetricCategories(Enum):
    residual = "Residual Diagnostics"
    entropy = "Information Entropy"
    class_err = "Forecast Errors - Classification"
    reg_err = "Forecast Errors - Regression"
    unknown = "Unknown"


class SampleTypes(Enum):
    insample = "insample"
    outofsample = "outofsample"


class Calculation(Enum):
    single = "single"
    rolling = "rolling"
    both = "both"


class SaveModes(Enum):
    svg = "svg"
    html = "html"
    text = "text"
    obj = "obj"
    minimal = "minimal"


class PathConst:
    default_eval_output_path: Path = Path("output/")
    default_analyse_output_path: Path = Path("analyse/")
    default_save_output_path: Path = Path("scorecards/")
    html_report_template_url: Path = Path("library/pdf_layouts/scorecard/report.html")
    css_report_template_url: Path = Path("library/pdf_layouts/scorecard/report.css")


class ComputationalComplexity(Enum):
    low = "low"
    medium = "medium"
    high = "high"


@dataclass
class ScoreCardMetadata:
    save_path: Path
    project_name: str = ""
    project_description: str = ""
    model_name: str = ""
    model_description: str = ""
    dataset_name: str = ""
    dataset_description: str = ""

    def __setattr__(self, key: str, item: Any) -> None:
        if key == "project_name":
            self.save_path = Path(
                os.path.join(PathConst.default_eval_output_path, item)
            )

        super().__setattr__(key, item)


class PrintMode(Enum):
    extended = "extended"
    minimal = "minimal"
    minimal_table = "minimal_table"

    @staticmethod
    def from_str(value: Union[str, "PrintMode"]) -> "PrintMode":
        if isinstance(value, PrintMode):
            return value
        for strategy in PrintMode:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown PrintMode: {value}")


class NamingPrefixes:
    model = "Model_"
    dataset = "Dataset_"
    project = "Project_"



from typing import Any, Tuple, Union

import numpy as np
import pandas as pd
from statsmodels.stats.diagnostic import acorr_ljungbox

from krisi.evaluate.type import Predictions, ProbabilitiesDS, TargetsDS


def ljung_box(
    y: TargetsDS, predictions: Predictions
) -> Union[pd.DataFrame, Tuple[Any, Any], Tuple[Any, Any, Any, Any]]:
    return acorr_ljungbox(predictions)


def create_one_hot(ds: Union[TargetsDS, ProbabilitiesDS]) -> pd.DataFrame:
    return pd.get_dummies(ds)


def brier_multi(targets: TargetsDS, probs: ProbabilitiesDS, **kwargs):
    # See https://stats.stackexchange.com/questions/403544/how-to-compute-the-brier-score-for-more-than-two-classes
    return np.mean(np.sum((probs - create_one_hot(targets)) ** 2, axis=1))



from dataclasses import dataclass
from typing import Any, Dict, Generic, List

from .metric import Metric
from .type import (
    MetricFunction,
    MetricResult,
    PredictionsDS,
    ProbabilitiesDS,
    TargetsDS,
)


@dataclass
class Group(Metric, Generic[MetricResult]):
    def __init__(
        self, name: str, key: str, metrics: List[Metric], func: MetricFunction
    ) -> None:
        self.metrics = metrics
        self.group_func = func
        self.key = key
        self.name = name

    def evaluate(
        self, y: TargetsDS, predictions: PredictionsDS, probabilities: ProbabilitiesDS
    ) -> List[Metric]:
        if self.accepts_probabilities and probabilities is not None:
            results = self.group_func(y, predictions, probabilities)
        else:
            results = self.group_func(y, predictions)
        return [metric._evaluation(results) for metric in self.metrics]

    def evaluate_over_time(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDS,
        rolling_args: Dict[str, Any],
    ) -> List[Metric]:
        if self.accepts_probabilities and probabilities is not None:
            results = self.group_func(y, predictions, probabilities)
        else:
            results = self.group_func(y, predictions)
        return [
            metric._rolling_evaluation(results, rolling_args=rolling_args)
            for metric in self.metrics
        ]

    def __str__(self) -> str:
        return "\n".join([metric.__str__() for metric in self.metrics])

    def __repr__(self) -> str:
        return "\n".join([metric.__repr__() for metric in self.metrics])



import logging
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, List, Optional, Tuple, Union

import pandas as pd

from krisi.evaluate.type import (
    ComputationalComplexity,
    MetricCategories,
    MetricFunction,
    MetricResult,
    PredictionsDS,
    ProbabilitiesDS,
    SampleTypes,
    TargetsDS,
)
from krisi.report.console import print_metric
from krisi.report.type import InteractiveFigure, PlotFunction, plotly_interactive
from krisi.utils.iterable_helpers import isiterable, string_to_id


@dataclass
class Metric(Generic[MetricResult]):
    """
    Class representing a metric.

    Parameters
    ----------
    name: str
        The name of the metric.
    key: str
        The key used to reference the metric.
    category: MetricCategories
        The category of the metric.
    result: Optional[Union[Exception, MetricResult, List[MetricResult]]]
        The result of the evaluated `Metric`, by default None
    result_rolling: Optional[Union[Exception, MetricResult, List[MetricResult]]]
        The result of the evaluated `Metric` over time, by default None
    parameters: dict
        The paramaters that are passed into the evaluation function (param: `func`), by default field(default_factory=dict)
    func: Callable
        The function used to compute the metric.
    plot_funcs: List[Callable]
        List of functions used to plot the metric.
    plot_func_rolling: Callable
        Function used to plot the rolling metric value.
    info: str
        Additional information about the metric.
    restrict_to_sample: Optional[SampleTypes]
        Weather the `Metric` should only be evaluated on insample or out of sample data, by default None
    comp_complexity: Optional[ComputationalComplexity]
        How resource intensive the calculation is, by default None
    disable_rolling: bool
        Whether the metric should not be evaluated on a rolling basis, by default False
        If this is switched on, the metric will still be evaluated when `evaluate_over_time` is called
        but not on a rolling basis.
    accepts_probabilities: bool
        Whether the metric accepts probabilities as input, by default False
    supports_multiclass: bool
        Whether the metric supports multiclass classification, by default False
    """

    name: str
    key: str = ""
    category: Optional[MetricCategories] = None
    result: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    result_rolling: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    parameters: dict = field(default_factory=dict)
    func: Optional[MetricFunction] = None
    plot_funcs: Optional[List[Tuple[PlotFunction, Optional[Dict[str, Any]]]]] = None
    plot_func_rolling: Optional[Tuple[PlotFunction, Optional[Dict[str, Any]]]] = None
    info: str = ""
    restrict_to_sample: Optional[SampleTypes] = None
    comp_complexity: Optional[ComputationalComplexity] = None
    disable_rolling: bool = False
    accepts_probabilities: bool = False
    supports_multiclass: bool = False
    _from_group: bool = False

    def __post_init__(self):
        if self.key == "":
            self.key = string_to_id(self.name)

    def __setitem__(self, key: str, item: Any) -> None:
        setattr(self, key, item)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key, "Unknown Field")

    def __str__(self) -> str:
        return print_metric(self)

    def __repr__(self) -> str:
        print(print_metric(self, repr=True))
        return super().__repr__()

    def _evaluation(self, *args) -> "Metric":
        if self._from_group:
            return self
        try:
            result = self.func(*args, **self.parameters)
        except Exception as e:
            result = e
        self.__safe_set(result, key="result")
        return self

    def evaluate(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: Optional[ProbabilitiesDS] = None,
    ) -> None:
        assert (
            self.func is not None
        ), "`func` has to be set on Metric to calculate result."
        if self.accepts_probabilities and probabilities is not None:
            self._evaluation(y, predictions, probabilities)
        else:
            self._evaluation(y, predictions)

    def _rolling_evaluation(self, *args, rolling_args: dict) -> "Metric":
        if self._from_group:
            return self
        if self.disable_rolling:
            return self
        else:
            _df = pd.concat(args, axis="columns")
            try:
                df_rolled = (
                    _df.expanding()
                    if rolling_args["window"] is None
                    else _df.rolling(**rolling_args)
                )

                result_rolling = [
                    self.func(*single_window.values.T, **self.parameters)
                    for single_window in df_rolled
                ]
            except Exception as e:
                result_rolling = e
            self.__safe_set(result_rolling, key="result_rolling")

        return self

    def evaluate_over_time(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: Optional[ProbabilitiesDS] = None,
        rolling_args: dict = dict(),
    ) -> None:
        if self.accepts_probabilities and probabilities is not None:
            self._rolling_evaluation(
                y, predictions, probabilities, rolling_args=rolling_args
            )
        else:
            self._rolling_evaluation(y, predictions, rolling_args=rolling_args)

    def is_evaluated(self, rolling: bool = False):
        if rolling:
            return self.result_rolling is not None
        else:
            return self.result is not None

    def get_diagram_over_time(self) -> Optional[List[InteractiveFigure]]:
        return create_diagram_rolling(self)

    def get_diagrams(
        self,
    ) -> Optional[List[InteractiveFigure]]:
        return create_diagram(self)

    def print(
        self,
        mode: Union[str, List[str]] = "dummy",
        with_info: bool = False,
        input_analysis: bool = True,
        title: Optional[str] = None,
    ) -> None:
        pass

    def __safe_set(
        self, result: Union[Exception, MetricResult, List[MetricResult]], key: str
    ):
        if self.__dict__[key] is not None:
            raise ValueError("This metric already contains a result.")
        else:
            self.__dict__[key] = result


def create_diagram_rolling(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_func_rolling is None:
        logging.info("No plot_func_rolling (Plotting Function Rolling) specified")
        return None
    elif isinstance(obj.result_rolling, Exception) or obj.result_rolling is None:
        return None
    elif isiterable(obj.result_rolling):
        return [
            InteractiveFigure(
                f"{obj.key}_{obj.plot_func_rolling[0].__name__}",
                get_figure=plotly_interactive(
                    obj.plot_func_rolling[0], obj.result_rolling, title=obj.name
                ),
                title=f"{obj.name} - {obj.plot_func_rolling[0].__name__}",
                category=obj.category,
                plot_args=merge_default_dict(obj.plot_func_rolling[1]),
            )
        ]
    else:
        return None


def merge_default_dict(
    new_dict: Optional[Dict[str, Any]] = None,
    default_dict: Dict[str, Any] = dict(width=900.0, height=600.0),
) -> Dict[str, Any]:
    if new_dict is None:
        return default_dict
    else:
        return {**default_dict, **new_dict}


def create_diagram(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_funcs is None:
        logging.info("No plot_func (Plotting Function) specified")
        return None
    elif isinstance(obj.result, Exception) or obj.result is None:
        return None
    else:
        return [
            InteractiveFigure(
                f"{obj.key}_{plot_func.__name__}",
                get_figure=plotly_interactive(plot_func, obj.result, title=obj.name),
                title=f"{obj.name} - {plot_func.__name__}",
                category=obj.category,
                plot_args=merge_default_dict(plot_args),
            )
            for plot_func, plot_args in obj.plot_funcs
        ]



import os
from pathlib import Path
from typing import TYPE_CHECKING, List, Union

import pandas as pd
from rich.console import Console

from krisi.evaluate.type import MetricCategories, PathConst, SaveModes
from krisi.report.console import get_minimal_summary, get_summary

if TYPE_CHECKING:
    from krisi.evaluate.scorecard import ScoreCard


def save_object(obj: "ScoreCard", path: Path) -> None:
    import dill

    final_path = Path(os.path.join(path, Path("scorecard.pickle")))

    with open(final_path, "wb") as file:
        dill.dump(obj, file)


def save_console(
    obj: "ScoreCard",
    path: Path,
    with_info: bool,
    save_modes: List[Union[SaveModes, str]],
) -> None:
    summary = get_summary(
        obj,
        repr=True,
        categories=[el.value for el in MetricCategories],
        with_info=with_info,
    )

    console = Console(record=True, width=120)
    with console.capture():
        console.print(summary)

    if SaveModes.text in save_modes or SaveModes.text.value in save_modes:
        console.save_text(os.path.join(path, Path("console.txt")), clear=False)
    if SaveModes.html in save_modes or SaveModes.html.value in save_modes:
        console.save_html(os.path.join(path, Path("console.html")), clear=False)
    if SaveModes.svg in save_modes or SaveModes.svg.value in save_modes:
        console.save_svg(
            os.path.join(path, Path("console.svg")),
            title="save_table_svg.py",
            clear=False,
        )


def save_minimal_summary(
    obj: "ScoreCard", path: Path, frame_or_series: bool = True
) -> None:
    text_summary = get_minimal_summary(obj, dataframe=frame_or_series)

    if isinstance(text_summary, (pd.Series, pd.DataFrame)):
        text_summary = text_summary.to_string()

    final_path = Path(os.path.join(path, Path("minimal.txt")))

    with open(final_path, "w", encoding="utf-8") as f:
        f.write(text_summary)


def load_scorecards(
    project_name: str, path: Union[str, Path] = PathConst.default_eval_output_path
) -> List["ScoreCard"]:
    import os
    import pickle

    if isinstance(path, str):
        path = Path(path)

    path = Path(
        os.path.join(
            path, os.path.join(Path(project_name), PathConst.default_save_output_path)
        )
    )
    project_files = os.listdir(path)

    scorecard_dirs = [
        directory
        for directory in project_files
        if os.path.isdir(os.path.join(path, directory))
    ]

    loaded_scorecards = []
    for scorecard_dir in scorecard_dirs:
        files = os.listdir(os.path.join(path, scorecard_dir))
        for file in files:
            if file.split(".")[-1] != "pickle":
                continue
            with open(
                Path(os.path.join(path, Path(f"{scorecard_dir}/{file}"))), "rb"
            ) as f:
                loaded_scorecards.append(pickle.load(f))

    return loaded_scorecards


def ensure_path(path: Union[str, Path]) -> None:
    if not os.path.exists(path):
        os.makedirs(path)



import pkgutil
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Union

import pandas as pd

from krisi.report.type import InteractiveFigure
from krisi.utils.iterable_helpers import wrap_in_list

if TYPE_CHECKING:
    import plotly.graph_objects as go


class VizualisationMethod(Enum):
    overlap = "overlap"
    seperate = "seperate"

    @staticmethod
    def from_str(value: Union[str, "VizualisationMethod"]) -> "VizualisationMethod":
        if isinstance(value, VizualisationMethod):
            return value
        for strategy in VizualisationMethod:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown VizualisationMethod: {value}")


def __calc_plot_names(
    df: pd.DataFrame,
    modes: List[VizualisationMethod],
    y_separate: bool,
    joint_plot_name: str,
    target_plot_name: str,
) -> List[str]:
    name_of_plots = [] + [target_plot_name] if y_separate else []

    for mode in modes:
        if mode == VizualisationMethod.seperate:
            name_of_plots += list(df.columns)

        if mode == VizualisationMethod.overlap:
            name_of_plots += [joint_plot_name]

    return name_of_plots


def __vizualise_with_plotly(
    y: pd.Series,
    preds: Union[pd.Series, List[pd.Series], pd.DataFrame],
    title: Optional[str],
    x_name: str,
    y_name: str,
    modes: List[VizualisationMethod],
    y_separate: bool,
    target_opacity: float,
    predictions_opacity: float,
    joint_plot_name: str,
    target_plot_name: str,
):
    import plotly as plt
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    df = (
        pd.concat(preds, axis="columns")
        if isinstance(preds, List)
        else (preds.to_frame() if isinstance(preds, pd.Series) else preds)
    )
    df = df.reindex(y.index.union(df.index))

    name_of_plots = __calc_plot_names(
        df, modes, y_separate, joint_plot_name, target_plot_name
    )
    num_plots = len(name_of_plots)

    fig = make_subplots(
        rows=num_plots,
        cols=1,
        shared_xaxes=True,
        subplot_titles=name_of_plots,
        horizontal_spacing=0.05,
        vertical_spacing=0.10,
    )
    y_trace = go.Scatter(
        x=df.index,
        y=y,
        name="y",
        line_color="red",
        legendwidth=0.0,
        legendgroup="y",
        showlegend=True,
        opacity=target_opacity,
    )
    traces = [
        go.Scatter(
            x=df.index,
            y=df[column],
            name=column,
            legendgroup="models",
            opacity=predictions_opacity,
            line_color=plt.colors.DEFAULT_PLOTLY_COLORS[i],
        )
        for i, column in enumerate(df.columns)
    ]

    if y_separate:
        y_trace.showlegend = False
        fig.add_trace(
            y_trace,
            row=1,
            col=1,
        )

    for mode in modes:
        if mode == VizualisationMethod.seperate:
            for i, column in enumerate(df.columns):
                fig.add_trace(
                    traces[i],
                    row=i + (2 if y_separate else 1),
                    col=1,
                )
                if not y_separate:
                    if i == 0:
                        y_trace.showlegend = True
                    else:
                        y_trace.showlegend = False
                    fig.add_trace(
                        y_trace,
                        row=i + 1,
                        col=1,
                    )

        if mode == VizualisationMethod.overlap:
            if not y_separate:
                if num_plots == 1:
                    y_trace.showlegend = True
                fig.add_trace(
                    y_trace,
                    row=num_plots,
                    col=1,
                )
            for i, column in enumerate(df.columns):
                fig.add_trace(
                    traces[i],
                    row=num_plots,
                    col=1,
                )

    for row in range(num_plots):
        fig.update_xaxes(title_text=x_name, row=row + 1, col=1)
        fig.update_yaxes(title_text=y_name, row=row + 1, col=1)
        fig.update_layout(**{f"xaxis{str(row+1)}_showticklabels": True})

    fig.update_layout(
        title=title,
        autosize=True,
        height=350.0 * num_plots,
        margin=dict(l=75, r=75, b=50, t=75, pad=0),
        hovermode="x unified",
    )
    fig.show()


def plot_y_predictions(
    y: pd.Series,
    preds: Union[List[pd.Series], pd.DataFrame],
    title: Optional[str] = "",
    x_name: str = "index",
    y_name: str = "value",
    mode: Union[str, VizualisationMethod, List[Union[str, VizualisationMethod]]] = [
        # VizualisationMethod.seperate,
        VizualisationMethod.overlap,
    ],
    y_separate: bool = False,
    target_opacity: float = 1.0,
    predictions_opacity: float = 0.75,
    joint_plot_name: str = "Joint Plot",
    target_plot_name: str = "Target (y)",
) -> None:
    modes = wrap_in_list(mode)
    if pkgutil.find_loader("plotly"):
        __vizualise_with_plotly(
            y,
            preds=preds,
            title=title,
            x_name=x_name,
            y_name=y_name,
            modes=[VizualisationMethod.from_str(mode) for mode in modes],
            y_separate=y_separate,
            target_opacity=target_opacity,
            predictions_opacity=predictions_opacity,
            joint_plot_name=joint_plot_name,
            target_plot_name=target_plot_name,
        )
    else:
        raise AssertionError(
            "`Plotly` is not installed. Install Plotly by running `pip install plotly` or unlock all reporting capability by `pip install krisi[plotting]`."
        )


def create_subplots_from_mutiple_plots(
    interactive_figures: List[InteractiveFigure], title: str
) -> "go.Figure":
    from plotly.subplots import make_subplots

    figures = [figure.get_figure(**figure.plot_args) for figure in interactive_figures]

    fig = make_subplots(
        rows=len(figures),
        cols=1,
        subplot_titles=[figure.title for figure in interactive_figures],
        row_heights=[500 for _ in range(len(figures))],
        specs=[[{"type": fig_.data[0].type}] for fig_ in figures],
    )

    for i, figure in enumerate(figures):
        fig.add_trace(figure.data[0], row=1 + i, col=1)

    fig.update_layout(height=500 * len(figures), title_text=title)

    return fig



import datetime
import os
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.evaluate.type import (
    NamingPrefixes,
    PathConst,
    Predictions,
    Probabilities,
    Targets,
)
from krisi.utils.iterable_helpers import is_int

if TYPE_CHECKING:
    from krisi.evaluate.scorecard import ScoreCard, ScoreCardMetadata


def handle_empty_metrics_to_display(
    scorecard: "ScoreCard", sort_by: Optional[str], metric_keys: Optional[List[str]]
) -> Tuple[List[str], Optional[str]]:
    if metric_keys is None:
        metric_keys = [
            metric.key
            for metric in scorecard.get_all_metrics(only_evaluated=True)
            if isinstance(metric.result, (float, int))
        ]

    if sort_by is not None:
        if sort_by not in metric_keys:
            metric_keys.insert(0, sort_by)
        else:
            metric_keys.remove(sort_by)
            metric_keys.insert(0, sort_by)

    else:
        sort_by = metric_keys[0]
    return metric_keys, sort_by


def handle_unnamed(
    y: Targets,
    predictions: Predictions,
    model_name: Optional[str],
    dataset_name: Optional[str],
    project_name: Optional[str],
) -> Tuple[str, str, str]:
    time = datetime.datetime.now()
    display_time = time.strftime("%Y%m%d-%H%M%S")

    if dataset_name is None:
        if isinstance(y, pd.Series) and y.name is not None:
            dataset_name = str(y.name)
        else:
            dataset_name = f"{NamingPrefixes.dataset}{display_time+str(uuid.uuid4()).split('-')[0]}"

    if model_name is None:
        if isinstance(predictions, pd.Series) and predictions.name is not None:
            model_name = str(predictions.name)
        else:
            model_name = (
                f"{NamingPrefixes.model}{display_time+str(uuid.uuid4()).split('-')[0]}"
            )

    if project_name is None:
        project_name = (
            f"{NamingPrefixes.dataset}{display_time+str(uuid.uuid4()).split('-')[0]}"
        )

    return model_name, dataset_name, project_name


def get_save_path(project_name: str) -> Path:
    return Path(os.path.join(PathConst.default_eval_output_path, project_name))


def last_model_name(metadata: "ScoreCardMetadata") -> Path:
    if metadata.model_name[:6] == NamingPrefixes.model and is_int(
        metadata.model_name[7]
    ):
        path = os.path.join(metadata.save_path, PathConst.default_save_output_path)
        if os.path.exists(path):
            listdir = os.listdir(
                os.path.join(metadata.save_path, PathConst.default_save_output_path)
            )
            dir_model_name = listdir[-1]
        else:
            dir_model_name = metadata.model_name
    else:
        dir_model_name = metadata.model_name

    return Path(dir_model_name)


def convert_to_series(
    data: Union[List[float], List[int], pd.Series, np.ndarray], name: str
) -> pd.Series:
    """Converts a list[floats or ints] or a numpy array to a pandas Series.

    Parameters
    ----------
    data : Union[List[float], List[int], pd.Series, np.ndarray]
        The data to convert.

    Returns
    -------
    pd.Series
        The converted data.
    """
    if isinstance(data, pd.Series):
        return data.rename(name)
    return pd.Series(data, name=name)


def ensure_df(data: Probabilities, name: str) -> pd.DataFrame:
    """Converts a Probabilities to a pandas DataFrame.

    Parameters
    ----------
    data : Probabilities
        The data to convert.

    Returns
    -------
    pd.DataFrame
        The converted data.
    """
    if isinstance(data, pd.DataFrame):
        return data
    elif isinstance(data, pd.Series):
        return data.to_frame(name)
    elif isinstance(data, np.ndarray):
        df = pd.DataFrame(data)
        df.index.name = name
        return df
    elif isinstance(data, list):
        df = pd.DataFrame(data)
        df.index.name = name
        return df
    else:
        raise ValueError(f"Data type {type(data)} not supported.")



from .default_metrics_classification import (
    all_classification_metrics,
    minimal_classification_metrics,
    multilabel_classification,
)
from .default_metrics_regression import (
    all_regression_metrics,
    low_computation_regression_mterics,
    minimal_regression_metrics,
)
